/**
    @file ft800_demo_board.c

    @brief Sample application to demonstrate an FT51 driving an FT800, 
          displaying sensor information from the FT51 on the FT800.
 
    @details Sample application to demonstrate an FT51 driving an FT800, 
    displaying sensor information from the FT51 on the FT800

    - To fit the code within the 16k flash of the FT51, some of the FT800 
    primitives have been replaced with modified versions specific to this 
    application. The result is that not only does the function occupy less 
    space, but the function call also uses less, a major saving where a 
    function is used within the code at numerous points.

    @version 1.0.0

**/
/*
 * ============================================================================
 * History
 * =======
 *
 * (C) Copyright 2014, Future Technology Devices International Ltd.
 * ============================================================================
 *
 * This source code ("the Software") is provided by Future Technology Devices
 * International Limited ("FTDI") subject to the licence terms set out
 * http://www.ftdichip.com/FTSourceCodeLicenceTerms.htm ("the Licence Terms").
 * You must read the Licence Terms before downloading or using the Software.
 * By installing or using the Software you agree to the Licence Terms. If you
 * do not agree to the Licence Terms then do not download or use the Software.
 *
 * Without prejudice to the Licence Terms, here is a summary of some of the key
 * terms of the Licence Terms (and in the event of any conflict between this
 * summary and the Licence Terms then the text of the Licence Terms will
 * prevail).
 *
 * The Software is provided "as is".
 * There are no warranties (or similar) in relation to the quality of the
 * Software. You use it at your own risk.
 * The Software should not be used in, or for, any medical device, system or
 * appliance. There are exclusions of FTDI liability for certain types of loss
 * such as: special loss or damage; incidental loss or damage; indirect or
 * consequential loss or damage; loss of income; loss of business; loss of
 * profits; loss of revenue; loss of contracts; business interruption; loss of
 * the use of money or anticipated savings; loss of information; loss of
 * opportunity; loss of goodwill or reputation; and/or loss of, damage to or
 * corruption of data.
 * There is a monetary cap on FTDI's liability.
 * The Software may have subsequently been amended by another user and then
 * distributed by that other user ("Adapted Software").  If so that user may
 * have additional licence terms that apply to those amendments. However, FTDI
 * has no liability in relation to those amendments.
 * ============================================================================
 */

#include "FT_Platform.h"

#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <stdio.h>

#include "ft51.h"
#include "ft51_general_config.h"
#include "ft51_interrupts.h"
#include "ft51_iomux.h"

#include "registers\ft51_registers.h"
#include "registers\ft51_aio_registers.h"
#include "registers\ft51_iocell_registers.h"
#include "registers\ft51_iomux_registers.h"
#include "ft51_i2c_master.h"

#include "adt7310.h"

#define SPI_CLOCK_RATE 12000

// ---------------------------------------------------------------------------------
// Various FT800 specific definitions and variables ...
// ---------------------------------------------------------------------------------
#define SECTION_BEGIN_BITMAPS BEGIN(BITMAPS)
#define SECTION_BEGIN_POINTS  BEGIN(POINTS)
#define SECTION_BEGIN_LINES   BEGIN(LINES)
#define SECTION_CLOSE_BITMAPS END()
#define SECTION_CLOSE_POINTS  END()
#define SECTION_CLOSE_LINES   END()

// Values for display resolution to support various display panels
#ifdef SAMAPP_DISPLAY_QVGA
// Values specific to QVGA LCD display
#define FT_DISP_WIDTH    320
#define FT_DISP_HEIGHT   240
#define FT_DISP_HCYCLE   408
#define FT_DISP_HOFFSET   70
#define FT_DISP_HSYNC0     0
#define FT_DISP_HSYNC1    10
#define FT_DISP_VCYCLE   263
#define FT_DISP_VOFFSET   13
#define FT_DISP_VSYNC0     0
#define FT_DISP_VSYNC1     2
#define FT_DISP_PCLK       8
#define FT_DISP_SWIZZLE    2
#define FT_DISP_PCLKPOL    0
#else
// Values specific to WQVGA LCD display (DEFAULT)
#define FT_DISP_WIDTH    480
#define FT_DISP_HEIGHT   272
#define FT_DISP_HCYCLE   548
#define FT_DISP_HOFFSET   43
#define FT_DISP_HSYNC0     0
#define FT_DISP_HSYNC1    41
#define FT_DISP_VCYCLE   292
#define FT_DISP_VOFFSET   12
#define FT_DISP_VSYNC0     0
#define FT_DISP_VSYNC1    10
#define FT_DISP_PCLK       5
#define FT_DISP_SWIZZLE    0
#define FT_DISP_PCLKPOL    1
#endif

Ft_Gpu_Hal_Context_t   host;
Ft_Gpu_Hal_Context_t * phost;

// Globals used for building list of display commands ...
ft_uint32_t * pCmdBuffer;
ft_uint32_t   Ft_CmdBuffer_Index = 0;
ft_uint8_t    Ft_CmdBuffer[FT_CMD_FIFO_SIZE];

// ---------------------------------------------------------------------------------
// Millisecond timer related definitions and variables
// ---------------------------------------------------------------------------------
#define MICROSECONDS_TO_TIMER_TICKS(a) (65536 - ((a) * (48 / 12)))
volatile ft_uint16_t ms_timer      = 0;
volatile ft_uint16_t timer_display = 0;
volatile ft_uint16_t timer_force   = 0;
volatile ft_uint16_t timer_pulse   = 0;
volatile ft_uint16_t timer_temp    = 0;

// ---------------------------------------------------------------------------------
// Various graphics used in rendering ...
// ---------------------------------------------------------------------------------

typedef struct SAMAPP_Bitmap_header
{
	ft_uint8_t Format;
	ft_uint8_t Width;
	ft_uint8_t Height;
	ft_uint8_t Stride;
	ft_int32_t Offset;
} SAMAPP_Bitmap_header_t;

// Header and data for rendering temperature display ...
__code static const SAMAPP_Bitmap_header_t TemperatureImageInfo[1] =
{
	// format,width,height,stride,arrayoffset
	{ L1, 16, 10, 2, 0 }
};

__code static const ft_prog_uchar8_t TemperatureImageData[20] =
{
	// 'resolution ', 16, 'x', 10, 'format ', 'L1', 'stride ', 2, ' total size ', 20
	255,128, // x.0 degrees LARGE
	 62,  0, // x.1 degrees SMALL
	 62,  0, // x.2 degrees SMALL
	 62,  0, // x.3 degrees SMALL
	 62,  0, // x.4 degrees SMALL
	127,  0, // x.5 degrees MEDIUM
	 62,  0, // x.6 degrees SMALL
	 62,  0, // x.7 degrees SMALL
	 62,  0, // x.8 degrees SMALL
	 62,  0  // x.9 degrees SMALL
};

// Header and data for rendering pulse grid display ...
__code static const SAMAPP_Bitmap_header_t PulseGridImageInfo[1] =
{
	// format,width,height,stride,arrayoffset
	{ L1, 64, 10, 8, 0 }
};

__code static const ft_prog_uchar8_t PulseGridImageData[80] =
{
	// 'resolution ', 64, 'x', 10, 'format ', 'L1', 'stride ', 8, ' total size ', 80
	255,255,255,255,255,255,255,255,
	128,128,128,128,128,128,128,128,
	128,128,128,128,128,128,128,128,
	128,128,128,128,128,128,128,128,
	128,128,128,128,128,128,128,128,
	128,128,128,128,128,128,128,128,
	128,128,128,128,128,128,128,128,
	128,128,128,128,128,128,128,128,
	128,128,128,128,128,128,128,128,
	128,128,128,128,128,128,128,128
};

// Header and data for rendering pulse heart display ...
__code static const SAMAPP_Bitmap_header_t HeartImageInfo[1] =
{
	// format,width,height,stride,arrayoffset
	{ L1,  9, 14, 2, 0 }
};

__code static const ft_prog_uchar8_t HeartImageData[28] =
{
	// 'resolution ', 9, 'x', 14, 'format ', 'L1', 'stride ', 2, ' total size ', 28
	// heart - filled
	119,  0,
	255,128,
	255,128,
	127,  0,
	 62,  0,
	 28,  0,
	  8,  0,

	// heart - outline
	119,  0,
	136,128,
	128,128,
	 65,  0,
	 34,  0,
	 20,  0,
	  8,  0
};

// Headers and data for rendering logo (lines and grid overlaid on circles) ...
__code static const SAMAPP_Bitmap_header_t LogoGridImageInfo[1] =
{
	// format,width,height,stride,arrayoffset
	{ L1, 8, 4, 1, 0 }
};

__code static const ft_prog_uchar8_t LogoGridImageData[4] =
{
	// 'resolution ', 8, 'x', 4, 'format ', 'L1', 'stride ', 1, ' total size ', 8
	255, 136, 136, 136
};

__code static const SAMAPP_Bitmap_header_t LogoLineImageInfo[1] =
{
	// format,width,height,stride,arrayoffset
	{ L1, 32, 1, 4, 0 }
};

__code static const ft_prog_uchar8_t LogoLineImageData[4] =
{
	// 'resolution ', 32, 'x', 1, 'format ', 'L1', 'stride ', 4, ' total size ', 4
	255, 255, 255, 255
};

__code static const  ft_uint8_t * convertType[] = {"C", "F", "K"};

__code static const ft_uint8_t  * scrollText = "FT51 + FT800 sensor demonstration, using 2xADC, 2xSPI and GPIO ... press SW4 button to cycle through Celcius, Fahrenheit and Kelvin scales for temperature ...";

// ---------------------------------------------------------------------------------
// Various definitions and variables ...
// ---------------------------------------------------------------------------------
#define TIMER_DISPLAY   20

#define LOGO_XCOORD     24
#define LOGO_YCOORD     24

#define SCROLL_XCRD      8 // Ensure multiple of 8
#define SCROLL_YCRD    257
#define SCROLL_XWDH    464 // Ensure multiple of 8
#define SCROLL_YWDH     13

ft_void_t Force_Heartrate_Temperature();

static volatile uint8_t  hwButtons = 0; // What switches have been pressed

// ---------------------------------------------------------------------------------
// Force sensor related definitions and variables
// ---------------------------------------------------------------------------------
#define LED_FORCE_ON()          P3 &= ~((1 << 1))
#define LED_FORCE_OFF()         P3 |=  ((1 << 1))
#define FORCE_TIMER              20

#define GAUGE_DIAL_RADIUS       100
#define GAUGE_MARK_RADIUS       GAUGE_DIAL_RADIUS + 10
#define GAUGE_RANGE             900
#define GAUGE_RANGE_STEPS        10
#define GAUGE_MINOR_DIVISIONS     5 // Each minor division is 20
#define GAUGE_MAJOR_DIVISIONS     9 // Each major division is 100 (max force of 900)

#define GAUGE_XCOORD            180
#define GAUGE_YCOORD            126
#define GAUGE_TITLE_XCOORD      GAUGE_XCOORD - 20
#define GAUGE_TITLE_YCOORD      GAUGE_YCOORD - 120
#define GAUGE_VALUE_XCOORD      GAUGE_XCOORD
#define GAUGE_VALUE_YCOORD      GAUGE_YCOORD

#define GAUGE_DRAW_DIVISIONS    OPT_NOBACK | OPT_NOPOINTER
#define GAUGE_DRAW_POINTER      OPT_NOBACK | OPT_NOTICKS
#define GAUGE_FONT              28
#define GAUGE_FONT_OPTIONS      OPT_CENTERX | OPT_CENTERY

volatile ft_uint16_t sampleValueForce;
volatile ft_uint8_t  sampleReadyForce;
static   ft_uint16_t renderValueForce = 0;

// ---------------------------------------------------------------------------------
// Temperature sensor related definitions and variables
// ---------------------------------------------------------------------------------
#define TEMP_TIMER    20

#define TEMP_XCOORD  310
#define TEMP_YCOORD   26
#define TEMP_YALIGN    4
#define TEMP_XWIDTH   60
#define TEMP_YWIDTH  200
#define TEMP_XCENTRE TEMP_XCOORD + (TEMP_XWIDTH / 2)
#define TEMP_YCENTRE TEMP_YCOORD + (TEMP_YWIDTH / 2)
#define TEMP_YBOTTOM TEMP_YCOORD +  TEMP_YWIDTH

#define BUTTON_MASK    0x08
#define BUTTON_PRESS   0x08
#define BUTTON_RELEASE 0x00

volatile ft_int32_t sampleValueTemp;
volatile ft_uint8_t sampleReadyTemp;
static   ft_int32_t renderValueTemp;
static   ft_uint8_t sampleConvertIt;

typedef enum
{
	CONVERT_CELSIUS = 0,
	CONVERT_FAHRENHEIT,
	CONVERT_KELVIN,
	CONVERT_TOTAL
} eCONVERT_TYPE;

// ---------------------------------------------------------------------------------
// Heart rate related definitions and variables
// ---------------------------------------------------------------------------------
#define LED_HR_ON()  P3 &= ~((1 << 0))
#define LED_HR_OFF() P3 |= ((1 << 0))

#define PULSE_TIMER      10 // Sample rate (in ms)
#define PULSE_SAMPLES  1500 // Total number of samples held (15 seconds)
#define PULSE_DISPLAY   200 // Total number of samples that can be displayed
#define PULSE_THRESHOLD  20 // Threshold value to screen out noisy samples
#define PULSE_IS_BEAT    15 // Number of threshold events close together marked as a heart beat

#define PULSE_XCOORD    406
#define PULSE_YCOORD     26
#define PULSE_XWIDTH     64
#define PULSE_YWIDTH    200 // Total number of samples that can be displayed at any one time
#define PULSE_XCENTRE   PULSE_XCOORD + (PULSE_XWIDTH / 2)

volatile uint16_t sampleValuePulse;
volatile uint8_t  sampleReadyPulse;
static   uint16_t pulseSamples[PULSE_SAMPLES];

static   uint16_t pulseIndexer = 0;  // Next index to save pulse value at (circular buffer)
static   uint8_t  pulseCounter = 0;  // Used to count how many samples in a row are over the threshold

// ---------------------------------------------------------------------------------
// FT800 routines (modified versions of library code to reduce code size)
// ---------------------------------------------------------------------------------
// Modified version of <<<  >>>
ft_void_t FT800_Write08(ft_uint32_t addr, ft_uint8_t data)
{
	Ft_Gpu_Hal_StartTransfer(phost, FT_GPU_WRITE, addr);
	Ft_Gpu_Hal_Transfer8(phost, data);
	Ft_Gpu_Hal_EndTransfer(phost);
}

// Modified version of <<<  >>>
ft_void_t FT800_Write16(ft_uint32_t addr, ft_uint16_t data)
{
	Ft_Gpu_Hal_StartTransfer(phost, FT_GPU_WRITE, addr);
	Ft_Gpu_Hal_Transfer16(phost, data);
	Ft_Gpu_Hal_EndTransfer(phost);
}

// Modified version of <<<  >>>
ft_void_t FT800_CmdBuffer_AddCmd(ft_uint32_t cmd)
{
	// Copy the command instruction into buffer
	pCmdBuffer  = (ft_uint32_t *)&Ft_CmdBuffer[Ft_CmdBuffer_Index];
	*pCmdBuffer = cmd;
	// Increment the command index
	Ft_CmdBuffer_Index += FT_CMD_SIZE;
}

// Modified version of <<<  >>>
ft_void_t FT800_CmdBuffer_AddStr(const ft_char8_t * text)
{
	// Length of text + 1 for the null termination ...
	ft_uint8_t   size = strlen(text) + 1;

	// Copy text into command buffer ...
	strcpy(&Ft_CmdBuffer[Ft_CmdBuffer_Index], text);

	// Increment command index and align to 4 byte boundary ...
	Ft_CmdBuffer_Index += ((size + 3) & ~3);


	// this version saves 85 bytes over the version above ...
	//ft_uint8_t   size = 0;
	//ft_uint8_t * copy = &Ft_CmdBuffer[Ft_CmdBuffer_Index];
    //
	// Copy text into command buffer ...
	//do
	//{
	//	*copy++ = *text++;
	//	size++;
	//} while (*copy != 0);
	//*copy = 0;
	//size++;
    //
	// Increment command index and align to 4 byte boundary ...
	//Ft_CmdBuffer_Index += ((size + 3) & ~3);
}





















// Modified version of <<<  >>>
ft_void_t FT800_CmdBuffer_Send()
{
	if (Ft_CmdBuffer_Index > 0)
	{
		ft_uint8_t * databuff = Ft_CmdBuffer;
		ft_uint16_t  datasize = Ft_CmdBuffer_Index;

//		if (datasize > Ft_Gpu_Cmdfifo_Freespace(phost))
//		{
//			datasize = Ft_Gpu_Cmdfifo_Freespace(phost);
//		}

		Ft_Gpu_Hal_CheckCmdBuffer(phost, datasize);
		Ft_Gpu_Hal_StartCmdTransfer(phost, FT_GPU_WRITE, datasize);

		while (datasize--)
		{
			Ft_Gpu_Hal_Transfer8(phost, *databuff);
			databuff++;
		}

		Ft_Gpu_Hal_EndTransfer(phost);
		Ft_Gpu_Hal_Updatecmdfifo(phost, Ft_CmdBuffer_Index);
		Ft_Gpu_Hal_WaitCmdfifo_empty(phost);

		Ft_CmdBuffer_Index = 0;
	}
}

// ---------------------------------------------------------------------------------
// Millisecond timer ISR
// ---------------------------------------------------------------------------------
void ms_timer_interrupt(const uint8_t flags)
{
    (void) flags; // Flags not currently used

    if (ms_timer)
    {
        ms_timer--;
    }

    if (timer_display)
    {
    	timer_display--;
    }

    if (timer_force)
    {
    	timer_force--;
    }

    if (timer_pulse)
    {
    	timer_pulse--;
    }

    if (timer_temp)
    {
    	timer_temp--;
    }

    // Reload the timer
    TH0 = MSB(MICROSECONDS_TO_TIMER_TICKS(1000));
    TL0 = LSB(MICROSECONDS_TO_TIMER_TICKS(1000));
}

// ---------------------------------------------------------------------------------
// Millisecond timer interrupt initialisation
// ---------------------------------------------------------------------------------
void ms_timer_initialise(void)
{
    // Register our own handler for interrupts from Timer 0
    interrupts_register(ms_timer_interrupt, interrupts_timer0);

    // Timer0 is controlled by TMOD bits 0 to 3, and TCON bits 4 to 5.
    TMOD &= 0xF0; // Clear Timer0 bits
    TMOD |= 0x01; // Put Timer0 in mode 1 (16 bit)

    // Set the count-up value so that it rolls over to 0 after 1 millisecond.
    TH0 = MSB(MICROSECONDS_TO_TIMER_TICKS(1000));
    TL0 = LSB(MICROSECONDS_TO_TIMER_TICKS(1000));

    TCON &= 0xCF; // Clear Timer0's Overflow and Run flags
    TCON |= 0x10; // Start Timer0 (set its Run flag)
}

// ---------------------------------------------------------------------------------
// ADC ISR to measure force applied to force sensor
// ---------------------------------------------------------------------------------
ft_void_t adc_interrupt(const uint8_t flags)
{
    uint8_t interrupt = 0;
    uint8_t sample_l  = 0x00;
    uint8_t sample_h  = 0x00;

    (void)flags;

    // Determine the interrupt source <FORCE> ...
    IO_REG_INTERRUPTS_READ(AIO_INTERRUPTS_0_7, interrupt);
    if (interrupt & MASK_AIO_0_INT)
    {
        IO_REG_INTERRUPTS_READ(AIO_0_ADC_DATA_L_1, sample_l);
        IO_REG_INTERRUPTS_READ(AIO_0_ADC_DATA_U_1, sample_h);

        sampleValueForce = ((sample_h << 8) | sample_l);
        sampleReadyForce = TRUE;

        // Clear ADC interrupt register bit ...
        IO_REG_INTERRUPTS_WRITE(AIO_INTERRUPTS_0_7, MASK_AIO_0_INT);
    }

    // Determine the interrupt source <HEART RATE> ...
    IO_REG_INTERRUPTS_READ(AIO_INTERRUPTS_8_15, interrupt);
    if (interrupt & MASK_AIO_10_INT)
    {
        IO_REG_INTERRUPTS_READ(AIO_10_ADC_DATA_L_1, sample_l);
        IO_REG_INTERRUPTS_READ(AIO_10_ADC_DATA_U_1, sample_h);

        sampleValuePulse = ((sample_h << 8) | sample_l);
        sampleReadyPulse = TRUE;

        // Clear ADC interrupt register bit ...
        IO_REG_INTERRUPTS_WRITE(AIO_INTERRUPTS_8_15, MASK_AIO_10_INT);
    }
}

// ---------------------------------------------------------------------------------
// ADC SETUP
// ---------------------------------------------------------------------------------
void adc_setup (void)
{
    // Enable AIO Cell Controller
    IO_REG_GENERAL_WRITE(AIO_CONTROL, MASK_AIO_DEV_EN);

    // Enable force ADC
    // Select mode. Setup AIO_0 to ADC
    IO_REG_GENERAL_WRITE(AIO_MODE_0, MASK_AIO_0_MODE_1);
    // Enable AIO_0 interrupt
    IO_REG_GENERAL_WRITE(AIO_INTERRUPT_ENABLES_0_7, MASK_AIO_0_IEN);

    // Enable pulse rate ADC
    // Select mode. Setup AIO_10 to ADC
    IO_REG_GENERAL_WRITE(AIO_MODE_2, MASK_AIO_10_MODE_1);
    // Enable AIO_8 interrupt
    IO_REG_GENERAL_WRITE(AIO_INTERRUPT_ENABLES_8_15, MASK_AIO_10_IEN);

    // Register for ADC interrupts
    interrupts_register(adc_interrupt, interrupts_IO_cell);
}

// ---------------------------------------------------------------------------------
// I/O INITIALISATION
// ---------------------------------------------------------------------------------
ft_void_t iomux_setup()
{
    // Switch setup ...
    //IOMUX_INPUT(DIO_0, IOMUX_IN_PORT0_1); // SW1 (bit 0)
    //IOMUX_INPUT(DIO_7, IOMUX_IN_PORT0_2); // SW2 (bit 1)
    //IOMUX_INPUT(AIO_7, IOMUX_IN_PORT0_3); // SW3 (bit 2)
    IOMUX_INPUT(AIO_8, IOMUX_IN_PORT0_4); // SW4 (bit 3)

    // SPI Master setup ...
    IOMUX_OUTPUT(DIO_6, IOMUX_OUT_SPI_MASTER_SS_N_0); // Slave select for LCD
    IOMUX_INPUT(DIO_4,  IOMUX_IN_SPI_MASTER_MISO);
    IOMUX_OUTPUT(DIO_3, IOMUX_OUT_SPI_MASTER_MOSI);
    IOMUX_OUTPUT(DIO_5, IOMUX_OUT_SPI_MASTER_SCLK);

    IOMUX_set_pad_config(DIO_6,
                        IOMUX_PAD_DRIVE_CURRENT_16MA,
                        IOMUX_PAD_TRIGGER_NORMAL,
                        IOMUX_PAD_SLEW_RATE_FAST,
                        IOMUX_PAD_PULL_NONE);
    IOMUX_set_pad_config(DIO_4,
                        IOMUX_PAD_DRIVE_CURRENT_16MA,
                        IOMUX_PAD_TRIGGER_NORMAL,
                        IOMUX_PAD_SLEW_RATE_FAST,
                        IOMUX_PAD_PULL_NONE);
    IOMUX_set_pad_config(DIO_3,
                        IOMUX_PAD_DRIVE_CURRENT_16MA,
                        IOMUX_PAD_TRIGGER_NORMAL,
                        IOMUX_PAD_SLEW_RATE_FAST,
                        IOMUX_PAD_PULL_NONE);
    IOMUX_set_pad_config(DIO_5,
                        IOMUX_PAD_DRIVE_CURRENT_16MA,
                        IOMUX_PAD_TRIGGER_NORMAL,
                        IOMUX_PAD_SLEW_RATE_FAST,
                        IOMUX_PAD_PULL_NONE);

    // LED setup ...
    IOMUX_OUTPUT(AIO_11, IOMUX_OUT_PORT3_0); // Pulse rate
    IOMUX_OUTPUT(DIO_1,  IOMUX_OUT_PORT3_1); // Force
}

ft_void_t Ft_BootupConfig()
{
    // Boot up for FT800 followed by graphics primitive sample cases
    // Initial boot up DL - make the back ground green color

	ft_uint8_t FT_DLCODE_BOOTUP[12] =
    {
      240,240,240,  2, // GPU instruction CLEAR_COLOR_RGB
        7,  0,  0, 38, // GPU instruction CLEAR
        0,  0,  0,  0, // GPU instruction DISPLAY
    };

	// Do a power cycle for safer side
	Ft_Gpu_Hal_Powercycle(phost, FT_TRUE);
	Ft_Gpu_Hal_Rd16(phost, RAM_G);
	// Set the clk to external clock
	Ft_Gpu_HostCommand(phost, FT_GPU_EXTERNAL_OSC);
	Ft_Gpu_Hal_Sleep(10);

	// Switch PLL output to 48MHz
	Ft_Gpu_HostCommand(phost, FT_GPU_PLL_48M);
	Ft_Gpu_Hal_Sleep(10);

	// Do a core reset for safer side
	Ft_Gpu_HostCommand(phost, FT_GPU_CORE_RESET);

	// Access address 0 to wake up the FT800
	Ft_Gpu_HostCommand(phost, FT_GPU_ACTIVE_M);
	
	FT800_Write08(REG_GPIO_DIR,0x80 | Ft_Gpu_Hal_Rd8(phost, REG_GPIO_DIR));
	FT800_Write08(REG_GPIO,    0x80 | Ft_Gpu_Hal_Rd8(phost, REG_GPIO));

	{
		ft_uint8_t chipid;
		// Read Register ID to check if FT800 is ready.
		chipid = Ft_Gpu_Hal_Rd8(phost, REG_ID);
		while(chipid != 0x7C)
			chipid = Ft_Gpu_Hal_Rd8(phost, REG_ID);
	}
	// Configuration of LCD display
	FT800_Write16(REG_HCYCLE,   FT_DISP_HCYCLE);
	FT800_Write16(REG_HOFFSET,  FT_DISP_HOFFSET);
	FT800_Write16(REG_HSYNC0,   FT_DISP_HSYNC0);
	FT800_Write16(REG_HSYNC1,   FT_DISP_HSYNC1);
	FT800_Write16(REG_VCYCLE,   FT_DISP_VCYCLE);
	FT800_Write16(REG_VOFFSET,  FT_DISP_VOFFSET);
	FT800_Write16(REG_VSYNC0,   FT_DISP_VSYNC0);
	FT800_Write16(REG_VSYNC1,   FT_DISP_VSYNC1);
	FT800_Write08(REG_SWIZZLE,  FT_DISP_SWIZZLE);
	FT800_Write08(REG_PCLK,     FT_DISP_PCLK); // after this display is visible on the LCD
	FT800_Write08(REG_PCLK_POL, FT_DISP_PCLKPOL);
	FT800_Write16(REG_HSIZE,    FT_DISP_WIDTH);
	FT800_Write16(REG_VSIZE,    FT_DISP_HEIGHT);

	// Touch configuration - configure the resistance value to 1200 - this
	// value is specific to customer requirement and derived by experiment
	FT800_Write16(REG_TOUCH_RZTHRESH, 1200);
	FT800_Write08(REG_GPIO_DIR,       0xff);
	FT800_Write08(REG_GPIO,           0xff);

	// It is optional to clear the screen here
	Ft_Gpu_Hal_WrMem(phost, RAM_DL, (ft_uint8_t *)FT_DLCODE_BOOTUP, sizeof(FT_DLCODE_BOOTUP));
	FT800_Write08(REG_DLSWAP, DLSWAP_FRAME);
}

// ---------------------------------------------------------------------------------
// RENDERING ROUTINES
// ---------------------------------------------------------------------------------
ft_void_t ColourWhite()
{
	FT800_CmdBuffer_AddCmd(COLOR_RGB(255,255,255));
}

ft_void_t ColourBlack()
{
	FT800_CmdBuffer_AddCmd(COLOR_RGB(  0,  0,  0));
}

ft_void_t ColourCyan()
{
	FT800_CmdBuffer_AddCmd(COLOR_RGB(  0,160,192));
}

ft_void_t ColourBlue()
{
	FT800_CmdBuffer_AddCmd(COLOR_RGB(  0,  0,128));
}

ft_void_t  ColourRed()
{
	FT800_CmdBuffer_AddCmd(COLOR_RGB(255,  0,  0));
}

ft_void_t ClearBackground(ft_uint32_t colour)
{
	FT800_CmdBuffer_AddCmd(colour);
	FT800_CmdBuffer_AddCmd(CLEAR(1,1,1));
}

ft_void_t ClipFullScreen()
{
	FT800_CmdBuffer_AddCmd(SCISSOR_XY(0,0) );
	FT800_CmdBuffer_AddCmd(SCISSOR_SIZE(480, 272) );
}

ft_void_t ClipTemperature()
{
	FT800_CmdBuffer_AddCmd(SCISSOR_XY(TEMP_XCOORD, TEMP_YCOORD) );
	FT800_CmdBuffer_AddCmd(SCISSOR_SIZE(TEMP_XWIDTH, TEMP_YWIDTH) );
}

ft_void_t ClipPulse()
{
	FT800_CmdBuffer_AddCmd(SCISSOR_XY(PULSE_XCOORD,PULSE_YCOORD) );
	FT800_CmdBuffer_AddCmd(SCISSOR_SIZE(PULSE_XWIDTH, PULSE_YWIDTH) );
}

ft_void_t VertexII(ft_uint16_t xcrd, ft_uint16_t ycrd, ft_uchar8_t hndl, ft_uchar8_t cell)
{
	FT800_CmdBuffer_AddCmd(VERTEX2II(xcrd, ycrd, hndl, cell) );
}

ft_void_t VertexIIF(ft_uint16_t xcrd, ft_uint16_t ycrd)
{
	FT800_CmdBuffer_AddCmd(VERTEX2F(xcrd, ycrd));
}

ft_void_t RenderString(ft_int16_t xcrd, ft_int16_t ycrd, ft_uint8_t font, ft_uint16_t opts, const ft_char8_t * text)
{
	FT800_CmdBuffer_AddCmd(CMD_TEXT);
	FT800_CmdBuffer_AddCmd((((ft_uint32_t)ycrd << 16)|(xcrd & 0xffff)));
	FT800_CmdBuffer_AddCmd((((ft_uint32_t)opts << 16)|(font & 0xffff)));
	FT800_CmdBuffer_AddStr(text);
}

ft_void_t RenderNumber(ft_int16_t xcrd, ft_int16_t ycrd, ft_uint8_t font, ft_uint16_t opts, ft_int32_t nmbr)
{
	FT800_CmdBuffer_AddCmd(CMD_NUMBER);
	FT800_CmdBuffer_AddCmd((((ft_uint32_t)ycrd << 16) | (xcrd & 0xffff)));
	FT800_CmdBuffer_AddCmd((((ft_uint32_t)opts << 16) | (font & 0xffff)));
	FT800_CmdBuffer_AddCmd(nmbr);
}

ft_void_t RenderForceGauge(ft_uint16_t options, ft_uint16_t value)
{
	FT800_CmdBuffer_AddCmd(CMD_GAUGE);
	FT800_CmdBuffer_AddCmd((((ft_uint32_t)GAUGE_YCOORD          << 16) | (GAUGE_XCOORD          & 0xffff)));
	FT800_CmdBuffer_AddCmd((((ft_uint32_t)options               << 16) | (GAUGE_MARK_RADIUS     & 0xffff)));
	FT800_CmdBuffer_AddCmd((((ft_uint32_t)GAUGE_MINOR_DIVISIONS << 16) | (GAUGE_MAJOR_DIVISIONS & 0xffff)));
	FT800_CmdBuffer_AddCmd((((ft_uint32_t)GAUGE_RANGE           << 16) | (value                 & 0xffff)));
}

ft_void_t DisplayListStartNew()
{
	// Start of new display list ...
	FT800_CmdBuffer_AddCmd(CMD_DLSTART);

	// Clear entire display (using colour specified) ...
	ClearBackground(CLEAR_COLOR_RGB(0,0,128));
}

ft_void_t DisplayListTransmit()
{
	// End of display list ...
	FT800_CmdBuffer_AddCmd(DISPLAY());

	// Swap between frames
	FT800_CmdBuffer_AddCmd(CMD_SWAP);

	// Download the commands into fifo
	FT800_CmdBuffer_Send();
}

ft_void_t RenderLogo()
{
	static ft_uint8_t xoff[10] = {12,8,4,0,0,0,0,4,8,12};
	static ft_uint8_t xcrd;
	static ft_uint8_t ycrd;
	static ft_uint8_t draw;
	static ft_uint8_t fade;

	// Set point size of circles ...
	FT800_CmdBuffer_AddCmd(SECTION_BEGIN_POINTS);
	FT800_CmdBuffer_AddCmd(POINT_SIZE(20*16));

	// Set clip rectangle to cut top/bottom edges of partial circles off ...
	FT800_CmdBuffer_AddCmd(SCISSOR_XY(LOGO_XCOORD - 24, LOGO_YCOORD - 18));
	FT800_CmdBuffer_AddCmd(SCISSOR_SIZE(80, 37) );

	// Render partial circles changing from white(ish) to red(ish) ...
	xcrd = LOGO_XCOORD;
	for (draw = 0; draw < 2; draw++)
	{
		for (fade = 0; fade < 10; fade++)
		{
			FT800_CmdBuffer_AddCmd(COLOR_RGB(255 - fade * 4, 192 - fade * 16, 192 - fade * 16));
			VertexII(xcrd, LOGO_YCOORD, 0, 0);
			xcrd++;
		}
	}

	// Reset clip rectangle to entire display ...
	ClipFullScreen();

	// Render blue circle for gap between partial/full circles ...
	ColourBlue();
	VertexII(xcrd, LOGO_YCOORD, 0, 0);

	// Render full red circle (represents the silicon wafer, colour taken from logo on website) ...
	FT800_CmdBuffer_AddCmd(COLOR_RGB(173,  0,  0));
	VertexII(xcrd + 2, LOGO_YCOORD, 0, 0);
	FT800_CmdBuffer_AddCmd(SECTION_CLOSE_POINTS);

	// Draw lines over partial circles and grid over full circle ...
	FT800_CmdBuffer_AddCmd(SECTION_BEGIN_BITMAPS);
	ColourBlue();
	for (ycrd = 0; ycrd < 10; ycrd++)
	{
		VertexII(LOGO_XCOORD - 20, LOGO_YCOORD + ycrd * 4 - 20, 3, 0); // line
		for (xcrd = 0; xcrd < 5; xcrd++)
    	{
	    	VertexII(LOGO_XCOORD + xcrd * 8 + 2 + xoff[ycrd], LOGO_YCOORD + ycrd * 4 - 20, 2, 0); // grid
    	}
	}

	// Render logo text ...
	ColourWhite();
	RenderString(LOGO_XCOORD+46, LOGO_YCOORD - 23, 29, 0, "FTDI");
	RenderString(LOGO_XCOORD+46, LOGO_YCOORD,      27, 0, "Chip");
	FT800_CmdBuffer_AddCmd(SECTION_CLOSE_BITMAPS);
}

ft_void_t RenderForce()
{
	ft_uint8_t fade;
	ft_uint8_t size;

	sampleValueForce = (sampleValueForce / GAUGE_RANGE_STEPS) * GAUGE_RANGE_STEPS;

	if (sampleValueForce >= renderValueForce)
	{
		renderValueForce = sampleValueForce;
	}
	else
	{
		if (renderValueForce > GAUGE_RANGE_STEPS)
		{
			renderValueForce -= GAUGE_RANGE_STEPS;
		}
		else
		{
			renderValueForce = 0;
		}
	}

	// Render outer part of dial bezel ...
	FT800_CmdBuffer_AddCmd(SECTION_BEGIN_POINTS);
	size = GAUGE_DIAL_RADIUS;
	for (fade = 0; fade < 8; fade++)
	{
		FT800_CmdBuffer_AddCmd(POINT_SIZE(size*16));
		FT800_CmdBuffer_AddCmd(COLOR_RGB(fade * 32, fade * 32, fade * 32));
		VertexII(GAUGE_XCOORD, GAUGE_YCOORD, 0, 0);
		size--;
	}

	// Render inner part of dial bezel ...
	size -= 20;
	for (fade = 0; fade < 8; fade++)
	{
		FT800_CmdBuffer_AddCmd(POINT_SIZE(size*16));
		FT800_CmdBuffer_AddCmd(COLOR_RGB(255 - fade * 32, 255 - fade * 32, 255 - fade * 32));
		VertexII(GAUGE_XCOORD, GAUGE_YCOORD, 0, 0);
		size--;
	}

	// Render peak force by changing colour of space between bezel and pointer centre,
	// with more force = more red (gradually decays away to current force value)
	FT800_CmdBuffer_AddCmd(POINT_SIZE(size*16));
	FT800_CmdBuffer_AddCmd(COLOR_RGB((renderValueForce * 2) / 7, 0,0));
	VertexII(GAUGE_XCOORD, GAUGE_YCOORD, 0, 0);
	FT800_CmdBuffer_AddCmd(SECTION_CLOSE_POINTS);

	// Render title ...
	FT800_CmdBuffer_AddCmd(SECTION_BEGIN_BITMAPS);
	ColourWhite();
	RenderString(GAUGE_TITLE_XCOORD, GAUGE_TITLE_YCOORD, 26, 0, "Force");

	// Render divisions (no pointers) ...
	ColourBlack();
	RenderForceGauge(GAUGE_DRAW_DIVISIONS, sampleValueForce);

	// Render current force pointer ...
	FT800_CmdBuffer_AddCmd(COLOR_RGB(224,224,  0));
	RenderForceGauge(GAUGE_DRAW_POINTER, sampleValueForce);
	FT800_CmdBuffer_AddCmd(SECTION_CLOSE_BITMAPS);

	// Render pointer centre ...
	FT800_CmdBuffer_AddCmd(SECTION_BEGIN_POINTS);
	size -= 30;
	for (fade = 0; fade < 8; fade++)
	{
		FT800_CmdBuffer_AddCmd(POINT_SIZE(size*16));
		FT800_CmdBuffer_AddCmd(COLOR_RGB(fade * 32, fade * 32, fade * 32));
		VertexII(GAUGE_XCOORD, GAUGE_YCOORD, 0, 0);
		size--;
	}
	FT800_CmdBuffer_AddCmd(SECTION_CLOSE_POINTS);

	// Render current force value ...
	FT800_CmdBuffer_AddCmd(SECTION_BEGIN_BITMAPS);
	ColourBlack();
	RenderNumber(GAUGE_VALUE_XCOORD, GAUGE_VALUE_YCOORD, GAUGE_FONT, GAUGE_FONT_OPTIONS, sampleValueForce);
	FT800_CmdBuffer_AddCmd(SECTION_CLOSE_BITMAPS);
}

ft_void_t RenderTemperature()
{
	// Note for display, 1 degree = 100 lines, divisions shown every 0.1 degrees = 10 lines ...
	static ft_int32_t renderTrack;
	static ft_int32_t renderValue;
	static ft_int16_t renderYcord;
	static ft_uint8_t renderCount;
	static ft_uint8_t digitsWhole;
	static ft_uint8_t digitsTenth;
	static ft_uint8_t digitsFifth;

	// Displayed temperature tracks the sampled temperature, with speed of
	// tracking varying according to size of difference between the values ...
	renderTrack = (abs(sampleValueTemp - renderValueTemp) / 50) + 1;
	if (renderValueTemp < sampleValueTemp)
	{
		renderValueTemp += renderTrack;
	}
	else if (renderValueTemp > sampleValueTemp)
	{
		renderValueTemp -= renderTrack;
	}

	// Display window can show 2 degrees (with the centre point being the actual
	// temperature) and to ensure that divisions scroll smoothly onto display we
	// need to show one division extra at top/bottom of display, therefore add
	// 1.1 degrees to temperature for start of rendering and render	2.2 degrees
	// in total ... also convert to temperature scale selected ...
	if (sampleConvertIt == CONVERT_FAHRENHEIT)
	{
		renderValue = ((renderValueTemp * 9) / 5) + 3200 + 110;
	}
	else if (sampleConvertIt == CONVERT_KELVIN)
	{
		renderValue = renderValueTemp + 27315 + 110;
	}
	else
	{
		renderValue = renderValueTemp + 110;
	}

	// Calculate y offset and value to align rendering onto 0.1 degree boundries ...
	renderYcord = abs(renderValue) % 10;
	if (renderValue < 0)
	{
		renderValue += renderYcord - 10;
		renderYcord  = 10 - renderYcord + TEMP_YALIGN;
	}
	else
	{
		renderValue -= renderYcord;
		renderYcord += TEMP_YALIGN;
	}

	// set clip rectangle and display top half of gradient background ...
	ClipTemperature();
	FT800_CmdBuffer_AddCmd(CMD_GRADIENT);
	FT800_CmdBuffer_AddCmd((((ft_uint32_t)TEMP_YCOORD << 16) | (TEMP_XCOORD & 0xffff)));
	FT800_CmdBuffer_AddCmd(COLOR_RGB( 32, 32, 32));
	FT800_CmdBuffer_AddCmd((((ft_uint32_t)TEMP_YCENTRE << 16) | (TEMP_XCOORD & 0xffff)));
	ColourWhite();

	// Set clip rectangle and display bottom half of gradient background ...
	FT800_CmdBuffer_AddCmd(SCISSOR_XY(TEMP_XCOORD, TEMP_YCENTRE) );
	FT800_CmdBuffer_AddCmd(SCISSOR_SIZE(TEMP_XWIDTH, TEMP_YWIDTH / 2) );
	FT800_CmdBuffer_AddCmd(CMD_GRADIENT);
	FT800_CmdBuffer_AddCmd((((ft_uint32_t)TEMP_YCENTRE << 16) | (TEMP_XCOORD & 0xffff)));
	ColourWhite();
	FT800_CmdBuffer_AddCmd((((ft_uint32_t)TEMP_YBOTTOM << 16) | (TEMP_XCOORD & 0xffff)));
	FT800_CmdBuffer_AddCmd(COLOR_RGB( 32, 32, 32));

	// Set clip rectangle and display mid point line ...
	ClipTemperature();
	ColourCyan();
	VertexII(TEMP_XCOORD + 14, TEMP_YCENTRE, 3, 0); // line for mid-point

	// Render temperature scale ...
	for (renderCount = 0; renderCount < 22; renderCount++)
	{
		digitsWhole = abs(renderValue) % 100;
		digitsTenth = digitsWhole / 10;
		digitsFifth = digitsTenth %  2;

		if ((digitsWhole == 0) || (digitsFifth == 0))
		{
			if (renderValue <= 0)
			{
				ColourBlue(); // it's COLD
			}
			else
			{
				ColourRed(); // it's HOT
			}

			if (digitsWhole == 0)
			{
				// Render whole part of temperature using large font (every 1.0 degrees) ...
				RenderNumber(TEMP_XCOORD + 28, renderYcord, 28, OPT_CENTERX, abs(renderValue) / 100);
			}
			else
			{
				// Render fractional part of temperature using small font (every 0.2 degrees) ...
				RenderNumber(TEMP_XCOORD + 28, renderYcord+4, 26, OPT_CENTERX, digitsTenth);
			}
		}

		// Render division lines (large for 1.0, medium for 0.5, small for all other divisions) ...
		ColourBlack();
		VertexII(TEMP_XCOORD +  2, renderYcord + 12, 1, digitsTenth);
		VertexII(TEMP_XCOORD + 49, renderYcord + 12, 1, digitsTenth);

		// Next 0.1 degree division value to be displayed ...
		renderYcord += 10;
		renderValue -= 10;
	}

	// Reset clip rectangle to entire display ...
	ClipFullScreen();

	// Render title and temperature system displayed ...
	ColourWhite();
	RenderString(TEMP_XCENTRE, 6,                             26, OPT_CENTERX, "Temperature");
	RenderString(TEMP_XCENTRE, TEMP_YCOORD + TEMP_YWIDTH + 1, 28, OPT_CENTERX, convertType[sampleConvertIt]);
}

ft_void_t RenderPulse()
{
	// Notes for the Heart Rate Sensor ...
	//
	// 1)The ADC is 10 bits, therefore the maximum value that can be read is 1023.
	//
	// 2)Tapping the force sensor can induce noise in the heart rate sensor.
	//
	// 3)Due to the wildly varying data returned from the sensor, it can be difficult to write an
	//   algorithm which accurately detects heart beats for every user!
	//
	//   In this demonstration, we basically count the number of values over a specific threshold,
	//   only marking a heart beat as being detected when the count reaches a set level.
	//
	//   It should also be noted that the data suggests that we detect both the +ve and -ve parts
	//   of the pulse as 'detected beats', therefore every two 'detected beats' should be counted
	//   as one beat.
	//
	// 4)By using the VERTEX2F to plot lines and not multiplying the data returned from the sensor
	//   we only need an area 64 pixels wide to display the full range of the sensor (the VERTEX2F
	//   coordinates are specified in 1/16th of a pixel, 16 * 64 = 1024).
	//
	// 5)Only the last 200 (depth of display area) data points are plotted, with the width of each
	//   line drawn being the value read from the sensor, thereby allowing the user to see the raw
	//   data in real time (visualisation of the data may enable the user to develop a more accurate
	//   algorithm).

	ft_uint8_t  ycord;
	ft_uint16_t count;
	ft_uint16_t index;
	ft_uint8_t  total;
	ft_uint8_t  image;

	// Set clip rectangle for background and clear ...
	ClipPulse();
	ClearBackground(CLEAR_COLOR_RGB(255,255,255));

	// Draw background grid ...
	ColourCyan();

	ycord = PULSE_YCOORD + (pulseIndexer % 10) - 10;
	for (count = 0; count < 21; count++)
	{
		VertexII(PULSE_XCOORD, ycord, 0, 0);
		ycord += 10;
	}
	FT800_CmdBuffer_AddCmd(SECTION_CLOSE_BITMAPS);

	// Display the most recent samples (total samples displayed = PULSE_DISPLAY) ...
	FT800_CmdBuffer_AddCmd(SECTION_BEGIN_LINES);
	FT800_CmdBuffer_AddCmd(LINE_WIDTH(16));
	FT800_CmdBuffer_AddCmd(COLOR_A(64));
	ColourRed();

	index = pulseIndexer;
	ycord = PULSE_YCOORD;
	for (count = 0; count < PULSE_DISPLAY; count++)
	{
		// Next sample to process (handle wraparound of index) ...
		if (--index >= PULSE_SAMPLES)
		{
			index = PULSE_SAMPLES - 1;
		}

		VertexIIF( PULSE_XCOORD * 16, ycord*16);
		VertexIIF( PULSE_XCOORD * 16 + (pulseSamples[index] & 0x03FF), ycord*16);

		ycord++;
	}
	FT800_CmdBuffer_AddCmd(SECTION_CLOSE_LINES);

	// Display small heart symbols at heart beat detection points ...
	FT800_CmdBuffer_AddCmd(SECTION_BEGIN_BITMAPS);
	FT800_CmdBuffer_AddCmd(COLOR_A(128));
	ColourBlack();

	index = pulseIndexer;
	ycord = PULSE_YCOORD;
	image = 0;
	for (count = 0; count < PULSE_DISPLAY; count++)
	{
		// Next sample to process (handle wraparound of index) ...
		if (--index >= PULSE_SAMPLES)
		{
			index = PULSE_SAMPLES - 1;
		}

		// top bit set if heart beat has been detected ...
		if (pulseSamples[index] & 0x8000)
		{
			if (image)
			{
				VertexII(PULSE_XCOORD + 50, ycord, 4, 0);
			}
			else
			{
				VertexII(PULSE_XCOORD + 50, ycord, 4, 1);
			}

			image = (++image) % 2;
		}

		ycord++;
	}

	total = 0;
	for (count = 0; count < PULSE_SAMPLES; count++)
	{
		if (pulseSamples[count] & 0x8000)
		{
			total++;
		}
	}

	// Reset clip rectangle to entire display ...
	ClipFullScreen();
	FT800_CmdBuffer_AddCmd(COLOR_A(255));
	ColourWhite();
	RenderString(PULSE_XCENTRE,     6,                               26, OPT_CENTERX, "Pulse Rate");
	RenderString(PULSE_XCOORD,      PULSE_YCOORD + PULSE_YWIDTH + 6, 26, 0,           "BPM");
	RenderNumber(PULSE_XCOORD + 64, PULSE_YCOORD + PULSE_YWIDTH + 5, 27, OPT_RIGHTX, total * 2);
}

ft_void_t RenderScrollString()
{
	static ft_uint8_t  scrollChar = 0;
	static ft_uint16_t scrollXoff = SCROLL_XCRD + SCROLL_XWDH;

	// Set clip rectangle accordingly ...
	FT800_CmdBuffer_AddCmd(SCISSOR_XY(SCROLL_XCRD, SCROLL_YCRD));
	FT800_CmdBuffer_AddCmd(SCISSOR_SIZE(SCROLL_XWDH, SCROLL_YWDH));

	// Render next part of string ...
	RenderString(scrollXoff, SCROLL_YCRD-2,  18, 0, scrollText + scrollChar);

	// Reset clip rectangle to entire display ...
	ClipFullScreen();

	// Scrolling is performed by ...
	// 1) render text string moving in from right side until it hits the left side
	// 2) render text string at left side whilst advancing through the string character by character
	// 3) start at 1) again when last character reached.
	if (scrollXoff != SCROLL_XCRD - 6)
	{
		// Not reached left side yet so keep on going ...
		scrollXoff -= 2;
	}
	else
	{
		if ( *(scrollText + scrollChar) == 0 )
		{
			// Start again at the beginning if last character reached ...
			scrollChar = 0;
			scrollXoff = SCROLL_XCRD + SCROLL_XWDH;
		}
		else
		{
			// Next character to render string from ...
			scrollChar++;
			scrollXoff = SCROLL_XCRD;
		}
	}

	FT800_CmdBuffer_AddCmd(SECTION_CLOSE_BITMAPS);
}
// ---------------------------------------------------------------------------------
// MAIN ROUTINES
// ---------------------------------------------------------------------------------

ft_void_t setup()
{
	Ft_Gpu_HalInit_t halinit;

    // Initialise IO
    iomux_setup();

    // Set up ADC for force and pulse rate monitoring ...
    adc_setup();

	halinit.TotalChannelNum = 1;
              
	Ft_Gpu_Hal_Init(&halinit);
	host.hal_config.channel_no = 1;
	host.hal_config.spi_clockrate_khz = SPI_CLOCK_RATE;
    Ft_Gpu_Hal_Open(&host);
            
	phost = &host;

    Ft_BootupConfig();
    Ft_Gpu_Hal_Sleep(2000);//Show the booting up screen.

    Force_Heartrate_Temperature();

	// Close all the opened handles
    Ft_Gpu_Hal_Close(phost);
    Ft_Gpu_Hal_DeInit();
}

void loop()
{
}
// Nothing beyond this ... or is there?

ft_void_t CycleTemperatureScale()
{
	static ft_uint8_t checkButton = 0x08;

	// Check to see if button has been pressed to cycle temperature scale displayed ...
	hwButtons = ((~P0) >> 1) & 0x0F;

	if ((hwButtons & BUTTON_MASK) == checkButton)
	{
		if (checkButton == BUTTON_PRESS)
		{
			sampleConvertIt = (++sampleConvertIt) % CONVERT_TOTAL;
			checkButton = BUTTON_RELEASE;
		}
		else
		{
			checkButton = BUTTON_PRESS;
		}
	}
}

ft_int32_t StoreBitmaps(ft_int32_t storeat, SAMAPP_Bitmap_header_t * bitmap_info, ft_uint8_t * bitmap_data)
{
	// Calculate storage requirements for bitmap ...
	ft_int32_t bitmap_datasize = bitmap_info->Height * bitmap_info->Stride;

	// Copy raw data for bitmap into memory ...
	Ft_Gpu_Hal_WrMemFromFlash(phost, storeat, bitmap_data, bitmap_datasize);

	// Calculate and return address to store following data at...
	storeat += bitmap_datasize;
	return (storeat);
}

ft_void_t DlistBitmaps(ft_uint8_t handle, ft_int32_t storeat, SAMAPP_Bitmap_header_t * bitmap_info, ft_uint8_t cells)
{
	// Add bitmap handles to display list ...
	FT800_CmdBuffer_AddCmd(BITMAP_HANDLE(handle));
	FT800_CmdBuffer_AddCmd(BITMAP_SOURCE(storeat));
	FT800_CmdBuffer_AddCmd(BITMAP_SIZE(NEAREST, BORDER, BORDER, bitmap_info->Width, bitmap_info->Height));
	FT800_CmdBuffer_AddCmd(BITMAP_LAYOUT(bitmap_info->Format, bitmap_info->Stride, bitmap_info->Height / cells));
}

ft_void_t Force_Heartrate_Temperature()
{
	ft_uint16_t index  = 0;

    // Initialise the timer to provide millisecond count down timer(s) ...
    ms_timer_initialise();

	sampleValueForce = 0;
	sampleValueTemp  = 0;
	sampleValuePulse = 0;
	renderValueTemp  = 0;
	sampleConvertIt  = 0;

    // Turn the LEDs on ...
    P3 &= ~((1 << 0) | (1 << 1));

    ms_timer = 100;
    while(ms_timer);

    // Turn the LEDs off ...
    P3 |= ((1 << 0) | (1 << 1));

    // Initialise temperature ...
    temperature_reset();

    for (index = 0;index < PULSE_SAMPLES; index++)
    {
        pulseSamples[index] = 0;
    }

    // Initiate ADC Sample on AIO_0 - <FORCE>
    IO_REG_GENERAL_WRITE(AIO_SAMPLE_0, MASK_AIO_0_SAMPLE);

    // Initiate ADC Sample on AIO_8 <HEARTRATE>
    IO_REG_GENERAL_WRITE(AIO_SAMPLE_1, MASK_AIO_10_SAMPLE);

    // Copy bitmaps to FT800 ...
    StoreBitmaps(  0, PulseGridImageInfo,   PulseGridImageData);
    StoreBitmaps( 80, TemperatureImageInfo, TemperatureImageData);
    StoreBitmaps(100, LogoGridImageInfo,    LogoGridImageData);
    StoreBitmaps(104, LogoLineImageInfo,    LogoLineImageData);
    StoreBitmaps(108, HeartImageInfo,       HeartImageData);

    DisplayListStartNew();
    DlistBitmaps(0,   0, PulseGridImageInfo,    1);
    DlistBitmaps(1,  80, TemperatureImageInfo, 10);
    DlistBitmaps(2, 100, LogoGridImageInfo,     1);
    DlistBitmaps(3, 104, LogoLineImageInfo,     1);
    DlistBitmaps(4, 108, HeartImageInfo,        2);
	// Finish display list and send to FT800 ...
	DisplayListTransmit();


    // Main loop (read all measurements and render them on display)
	do
	{
    	// ----------------------------------------------------------------------
    	// SAMPLING
    	// ----------------------------------------------------------------------
		// <<< FORCE >>>
    	// Only request a sample after required number of milliseconds has passed ...
    	if (timer_force == 0)
    	{
	    	// Initiate ADC sample AIO_0 (to read force sensor)
	        IO_REG_GENERAL_WRITE(AIO_SAMPLE_0,  MASK_AIO_0_SAMPLE);
	        // Delay before initiating a sample request again ...
	    	timer_force = FORCE_TIMER;
    	}

    	// <<< PULSE RATE >>>
    	// Only request a sample after required number of milliseconds has passed ...
    	if (timer_pulse == 0)
    	{
        	// Initiate ADC Sample on AIO_8 <HEARTRATE>
            IO_REG_GENERAL_WRITE(AIO_SAMPLE_1, MASK_AIO_10_SAMPLE);
	        // delay before initiating a sample request again ...
	    	timer_pulse = PULSE_TIMER;
    	}

    	// <<< TEMPERATURE >>>
    	if (timer_temp == 0)
    	{
        	// Read the temperature from the SPI Master
    		sampleValueTemp = temperature_read();
    		timer_temp      = TEMP_TIMER;
    	}

    	// ----------------------------------------------------------------------
    	// PROCESSING OF SAMPLES
    	// ----------------------------------------------------------------------
    	// <<< FORCE >>>
		if (sampleReadyForce)
		{
			sampleReadyForce = FALSE;

			if (sampleValueForce > 900)
			{
				sampleValueForce = 900;
			}

	    	if (sampleValueForce > 10)
	    	{
	    		LED_FORCE_ON();
	    	}
	    	else
	    	{
	    		LED_FORCE_OFF();
	    	}
		}

    	// <<< PULSE RATE >>>
		if (sampleReadyPulse)
		{
    		sampleReadyPulse = FALSE;

    		// Record sample value ...
    		pulseSamples[pulseIndexer] = sampleValuePulse;

			// Keep track of how many samples in a block are greater than threshold ...
    		if (sampleValuePulse >= PULSE_THRESHOLD)
			{
				pulseCounter++;

				// Mark as a heart beat if number of samples in a block exceeds the set value ...
				if (pulseCounter == PULSE_IS_BEAT)
				{
					// Set top bit (bits 0-9 contain the actual sensor value)
					pulseSamples[pulseIndexer] |= 0x8000;
					// Start looking for the next heart beat ...
		        	pulseCounter                = 0;
					LED_HR_ON();
				}
			}
			else
			{
				LED_HR_OFF();
				// Deal with gaps (value < threshold) when counting number of samples in a block ...
				if (pulseCounter > 0)
				{
					pulseCounter--;
				}
			}

			pulseIndexer = (++pulseIndexer) % PULSE_SAMPLES;
		}

    	// <<<TEMPERATURE>>>
    	// Nothing to process ... it's hot or cold!!


    	// ----------------------------------------------------------------------
    	// RENDERING OF DISPLAY
    	// ----------------------------------------------------------------------
    	// Only update display after required number of milliseconds has passed ...
		if (timer_display == 0)
		{
			CycleTemperatureScale();

			DisplayListStartNew();
			RenderLogo();
			RenderForce();
	    	RenderTemperature();
	    	RenderPulse();
	    	RenderScrollString();
	    	DisplayListTransmit();

			// Reset timer for next display update ...
			timer_display = TIMER_DISPLAY;
		}
	} while (1);
}
